from math import isnan
class Histogram(object) :
    '''
    Simple 1-D histogram base class.

    You have to specify several things on construction.

    - xmin : Minimum x value
       * x < xmin goes into the underflow.
    - xmax : Maximum x value
       * x >= xmax goes into the overflow.        
    - nbins : Number of bins.
    - name : Name of this histogram.  This will be the key
             in the dictionary that's pickled.
    '''
    def __init__(self, xmin, xmax, nbins, name):
       
        self.xmin = xmin
        self.xmax = xmax
        self.name = name
        
        self.overflow = 0
        self.underflow = 0
        self.nan_count = 0
        self.bin_values = [0 for i in range(nbins)]

        # Inverse BinWidth, used in the calculation
        # of the index in the fill method
        self.__inv_bw = nbins/float(self.xmax - self.xmin) 
                       
    def fill(self, value):
        '''
        Note that the leading edge is inclusive
          * x < xmin goes into the underflow.
          * x >= xmax goes into the overflow.        
        '''
        if value >= self.xmin and value < self.xmax :
            index = int(self.__inv_bw * (value - self.xmin))
            self.bin_values[ index ] += 1
        else:
            if isnan(value):
                self.nan_count += 1
            else:
                if value >= self.xmax :
                    self.overflow += 1
                if value < self.xmin :
                    self.underflow += 1
                        
    def __getstate__(self):
        '''
        Allows Histograms to be pickled.
        '''
        state = {
            "name" : self.name,
            "xmin" : self.xmin,
            "xmax" : self.xmax,
            "overflow" : self.overflow,
            "underflow" : self.underflow,
            "nan_count" : self.nan_count,
            "bin_values" : self.bin_values
            }
        return state

    def __setstate__(self, state):
        '''
        Allows Histograms to be pickled.
        '''
        for name, value in state.items():
            setattr(self, name, value)

        nbins = len(self.bin_values)
        self.__inv_bw = nbins/float(self.xmax - self.xmin) 



#!/usr/bin/env python3

'''
Vanilla gaussian example.
'''


import numpy
import pylab  # type: ignore[import]
import scipy.stats  # type: ignore[import]

import voka.metrics

histograms = dict()
N_HISTOGRAMS = 1000
SCALE = 1.0
LOC = 0.0
SIZE = 1000
params = (SIZE/(SCALE * numpy.sqrt(2 * numpy.pi)), LOC, SCALE)
for i in range(N_HISTOGRAMS):
    data = numpy.random.normal(size=SIZE, loc=LOC, scale=SCALE)
    histograms['Histogram%d' % i] = numpy.histogram(data)

def gauss(x, *p):
    '''
    Gaussian distribution
    '''
    amplitude, mean, sigma = p
    return amplitude*numpy.exp(-(x-mean)**2/(2.*sigma**2))

test_stat = voka.metrics.norm_chisq
NDOF = None
T_dist = list()
for i in range(N_HISTOGRAMS):
    h = histograms["Histogram%d" % i]
    bin_values = h[0]
    bin_edges = h[1]
    bin_centers = (bin_edges[:-1] + bin_edges[1:])/2
    bw = bin_centers[1] - bin_centers[0]
    expectation = [gauss(x, bw*params[0], LOC, SCALE) for x in bin_centers]

    if not NDOF:
        NDOF = len(bin_values) - 2

    T_dist.append(test_stat(bin_values, expectation))

if NDOF is None:
    raise RuntimeError('NDOF was never set (is None)')

pylab.figure(1)
pylab.hist(T_dist, bins=100)

pylab.figure(2)
rv = scipy.stats.chi2(NDOF)
x_values = range(20)
pylab.plot(x_values, rv.pdf(x_values), 'k-')

pylab.figure(3)
h = histograms["Histogram0"]
bin_values = h[0]
bin_edges = h[1]
bin_centers = (bin_edges[:-1] + bin_edges[1:])/2
bw = bin_centers[1] - bin_centers[0]
print("bw = %f" % bw)
pylab.plot(bin_centers, bin_values)
pylab.plot(bin_centers, [gauss(bc, bw*params[0], LOC, SCALE)
                         for bc in bin_centers])

pylab.show()



#!/usr/bin/env python
'''
Tests the LOF function.
'''

import unittest

import random
import numpy

import voka.lof

class TestLOF(unittest.TestCase):
    '''
    Test the LOF function.  Generate dummy histograms
    and ensure the algorithm runs as expected.

    The LOF function takes a test point, a k-distance,
    and a cluster (i.e. benchmark).  For test 'point'
    and benchmark we use histograms here since that's
    the focus and main input of this project.
    '''
    def setUp(self):
        '''
        Generate a test point and benchmark set.
        Sampling from a Gaussian distribution.
        '''
        dist_mean = 0.
        sigma = 1.

        self.test_hist = numpy.histogram([random.gauss(dist_mean, sigma)
                                          for _ in range(1000)])[0]
        self.reference_collection = \
            [numpy.histogram([random.gauss(dist_mean, sigma)
                              for i in range(1000)])[0]
             for j in range(5)]

    def test_lof_basic_exection(self):
        '''
        Execute the fuction.
        TODO: Test other distributions and k-distances.
              Test for pathological inputs and ensure the fail
              in a sane manner.
        '''
        k_distance = 3
        result = voka.lof.local_outlier_factor(self.test_hist,
                                               k_distance,
                                               self.reference_collection)
        self.assertFalse(numpy.isnan(result))

if __name__ == '__main__':
    unittest.main()



'''
This module contains the compare function.
'''
import voka.metrics

ALL_METRICS = {"NormChiSq":voka.metrics.norm_chisq,
               "ShapeChiSq":voka.metrics.shape_chisq,
               "BDM":voka.metrics.bdm,
               "KolmogorovSmirnov":voka.metrics.kolmogorov_smirnov,
               "LLHRatio":voka.metrics.llh_ratio,
               "LLHValue":voka.metrics.llh_value,
               "CramerVonMises":voka.metrics.cramer_von_mises,
               "AndersonDarling":voka.metrics.anderson_darling}

DEFAULT_METRICS = {"ShapeChiSq":voka.metrics.shape_chisq,
                   "AndersonDarling":voka.metrics.anderson_darling}

def compare(values1, values2, metrics=None):
    r'''
    To use all metrics set metrics to voka.compare.ALL_METRICS,
    otherwise pass a dictionary where the values are the metrics
    to use with named keys (can be anything), e.g.:

        metrics = {
                   'foo': voka.metrics.norm_chisq, 
                   'bar': voka.metrics.anderson_darling
                  }

    Output:
        result : dict
            test name : value of test statistic
            Will be empty if no tests enabled, or histograms inconsistent.
    '''

    result = {}
    if len(values1) != len(values2):
        print("ERROR : sequences are inconsistent.")
        return result

    _metrics = metrics if metrics else DEFAULT_METRICS

    for name, metric in _metrics.items():
        result[name] = metric(values1, values2)

    return result



#!/usr/bin/env python3

'''
Example illustrating the classic method, meaning fitting by hand.
'''

import numpy
import pylab  # type: ignore[import]
import scipy.optimize  # type: ignore[import]

# Histogram the arrival time, expected to be gaussian, of the charge
# sampled from a gaussian.
data = list()
pe_time_dist = numpy.random.normal(size=1000)
for time in pe_time_dist:
    # for each time the charge is sampled from
    # another distribution
    charge = numpy.random.normal(loc=5.0, scale=0.25)
    for c in range(int(charge)):
        data.append(time)

def gauss(x, *p):
    '''
    The gaussian distribution.
    '''
    amplitude, mean, sigma = p
    return amplitude*numpy.exp(-(x-mean)**2/(2.*sigma**2))

hist, bin_edges = numpy.histogram(data)
bin_centers = (bin_edges[:-1] + bin_edges[1:])/2
p0 = [1000., 0., 1.]
p1 = [1000., 0., 1.]
print(hist)
print(bin_centers)

popt, pcov = scipy.optimize.curve_fit(gauss, bin_centers, hist, p0=p1)
fitted_amplitude = popt[0]
fitted_mean = popt[1]
fitted_std_dev = popt[2]

print('Fitted amplitude = %f' % fitted_amplitude)
print('Fitted mean = %f' % fitted_mean)
print('Fitted standard deviation = %f' % fitted_std_dev)

hist_fit = gauss(bin_centers, *popt)
pylab.hist(data)
pylab.plot(bin_centers, gauss(bin_centers, *p0), label='Initial')
pylab.plot(bin_centers, hist_fit, label='Fitted data')

pylab.show()



"""Rendering histograms."""

import pylab  # type: ignore[import]

def draw(sample, title=None, color = "green", ylim = None, log = False, yerr = None):
    r"""Make a matplotlib figure of a sample.
    Arguments:
        title : Plot title.
        color : matplotlib color string
            Color of the histogram area.
        ylim : None or (float, float)
            Axis limits (ymin, ymax). If None, use 0.1 margin around data.
        log : bool
            Draw the y-axis in a logarithmic scale.
        yerr : None, float, or array/list of floats
            Error bars around the histogram.

    r"""

    if not ylim:
        if log:
            positive_y_values = [v for v in sample if v > 0]
            ymin = 0.9 * min(positive_y_values) if positive_y_values else 0.1
        else:
            ymin = 0.9 * min(sample)
        ymax = 1.1 * max(sample)
    else:
        ymin, ymax = ylim

    pylab.ylim(ymin, ymax)
    pylab.title(title)

    left_edges = list(range(len(sample)))

    pylab.bar(left_edges,
              sample,
              linewidth = 0,
              align = "edge",
              color = color,
              log = log,
              yerr = yerr)

def draw_comparisons(test_sample,
                     benchmark_samples,
                     title=None,
                     colors = {'test': 'blue',  'benchmarks': 'green'},
                     log = False):
    r"""Make a matplotlib figure of comparing test_histogram to benchmark_histogram.
        Arguments:
            draw_ratio : bool
                Draw a ratio test_histogram/benchmark_histogram in a third subplot,
                with error bars. Always linear y-scale. The histograms have to use
                the same binning!
            color : matplotlib color string
                Color of the histogram areas.
            log : bool
                Draw the y-axis in a logarithmic scale.

    r"""

    # Need to figure out how many subplots to make.
    # Also need to scale the title

    subplot_start = {2 : 211,
                     3 : 311,
                     4 : 221,
                     5 : 321,
                     6 : 321,
                     7 : 331,
                     8 : 331,
                     9 : 331,
    } # matplotlib chokes on 431 and 431

    n_plots = min(1 + len(benchmark_samples), 9)

    fig_idx = subplot_start[n_plots]
    ax = pylab.subplot(fig_idx)
    ax.set_title(title)
    draw(test_sample, color = colors['test'], log = log)

    samples = benchmark_samples[:8] \
        if len(benchmark_samples) > 8 \
           else benchmark_samples

    for sample in samples:
        fig_idx += 1
        pylab.subplot(fig_idx)
        draw(sample, color = colors['benchmarks'], log = log)

def draw_ratios(test_sample,
                benchmark_samples,
                title=None,
                colors = {'test': 'blue',  'benchmarks': 'green'},
                log = False):

    # Need to figure out how many subplots to make.
    # Also need to scale the title

    subplot_start = {2 : 211,
                     3 : 311,
                     4 : 221,
                     5 : 321,
                     6 : 321,
                     7 : 331,
                     8 : 331,
                     9 : 331,
    } # matplotlib chokes on 431 and 431

    n_plots = min(len(benchmark_samples), 9)

    samples = benchmark_samples[:8] \
        if len(benchmark_samples) > 8 \
           else benchmark_samples

    fig_idx = subplot_start[n_plots]
    for sample in samples:
        pylab.subplot(fig_idx)
        fig_idx += 1

        ratio = [t/s if s else 0. for t,s in zip(test_sample, sample)]

        draw(ratio, color = colors['benchmarks'], log = log)

def draw_ratio(test_sample,
               benchmark_sample,
               colors = {'test': 'blue',  'benchmarks': 'green'},
               log = False):
    r"""Make a matplotlib figure of comparing test_histogram to benchmark_histogram.
        Arguments:
            draw_ratio : bool
                Draw a ratio test_histogram/benchmark_histogram in a third subplot,
                with error bars. Always linear y-scale. The histograms have to use
                the same binning!
            color : matplotlib color string
                Color of the histogram areas.
            log : bool
                Draw the y-axis in a logarithmic scale.

    r"""

    # Need to figure out how many subplots to make.
    # Also need to scale the title

    ax = pylab.subplot(221)
    draw(test_sample, color = colors['test'], log = log)

    pylab.subplot(222)
    draw(benchmark_sample, color = colors['benchmarks'], log = log)

    ratio_sample = [t/b if b else 0.
                    for t,b in zip(test_sample, benchmark_sample)]
    pylab.subplot(223)
    draw(ratio_sample)



#!/usr/bin/env python3

'''
This example illustrates the failure of traditional statistical
comparisons, such as chi^2, for weighted histograms from
stochastic processes.

Histogram the arrival time, expected to be gaussian, of the charge
sampled from a gaussian.
'''

import numpy
import pylab  # type: ignore[import]
import scipy.stats  # type: ignore[import]

import voka.metrics

histograms = dict()
N_HISTOGRAMS = 100
for i in range(N_HISTOGRAMS):
    charge_data = list()
    time_data = numpy.random.normal(size=1000)
    for time in time_data:
        # for each time the charge is sampled from
        # another distribution
        charge = numpy.random.normal(loc=5.0, scale=0.25)
        for c in range(int(charge)):
            charge_data.append(time)
    histograms['ChargeHistogram%d' % i] = numpy.histogram(charge_data)
    histograms['TimeHistogram%d' % i] = numpy.histogram(time_data)

charge_T_dist = list()
time_T_dist = list()
test_stat = voka.metrics.norm_chisq

NDOF = None
for i in range(N_HISTOGRAMS):
    for j in range(i+1, N_HISTOGRAMS):
        ch1 = histograms["ChargeHistogram%d" % i][0]
        ch2 = histograms["ChargeHistogram%d" % j][0]
        th1 = histograms["TimeHistogram%d" % i][0]
        th2 = histograms["TimeHistogram%d" % j][0]

        if not NDOF:
            NDOF = len(ch1)

        charge_T_dist.append(test_stat(ch1, ch2))
        time_T_dist.append(test_stat(th1, th2))

if NDOF is None:
    raise RuntimeError('NDOF was never set (is None)')

rv = scipy.stats.chi2(NDOF)
pylab.figure(1)
pylab.hist(charge_T_dist, density=True, bins=100)

pylab.figure(2)
pylab.hist(time_T_dist, density=True, range=(0, 100), bins=100)
x = range(100)
pylab.plot(x, rv.pdf(x), 'k-')
pylab.show()



#!/usr/bin/env python
'''
Module that tests the individual metrics.
'''

import unittest

import random
from math import isnan

import numpy

import voka.compare
import voka.metrics

class TestMetrics(unittest.TestCase):
    '''
    Test class that sets the test conditions and
    tests all the metrics defined in voka.compare.ALL_METRICS
    '''
    def setUp(self):
        hist1 = numpy.histogram([random.gauss(0., 1.)
                                 for _ in range(1000)])
        hist2 = numpy.histogram([random.gauss(0., 1.)
                                 for _ in range(1000)])

        self.gaussian1 = hist1[0]
        self.gaussian2 = hist2[0]

    def test_all(self):
        for metric in voka.compare.ALL_METRICS.values():
            result = metric(self.gaussian1, self.gaussian2)
            self.assertFalse(isnan(result))

    def test_default_compare(self):
        result = voka.compare.compare(self.gaussian1, self.gaussian2)
        self.assertEqual(len(result), 2)
        self.assertTrue('AndersonDarling' in result)
        self.assertTrue('ShapeChiSq' in result)

if __name__ == '__main__':
    unittest.main()



from math import fabs, log, sqrt

from scipy.special import binom  # type: ignore[import]


def norm_chisq(vector1, vector2):
    r"""Compare sequences vector1, vector2 with a Chi^2 test.
        Assumes they have the same binning.
        Output:
        chisq : float
            A Chi^2 sum over all bins.
    """
    assert len(vector1) == len(vector2)

    # calculate the \chi^2 test statistic
    terms = [(u - v)**2/(u + v) \
             for u, v in zip(vector1, vector2)\
             if u > 0 and v > 0]
    result = sum(terms)
    return result

def shape_chisq(vector1, vector2):
    r"""Compare sequences vector1, vector2 with a Chi^2 test after
    normalizing them.
    Output:
        chisq : float
            A Chi^2 sum over all bins.
    """
    assert len(vector1) == len(vector2)
    
    sum1 = sum(vector1)
    sum2 = sum(vector2)
    try:
        n1sq = sum1**2
        n2sq = sum2**2
        terms = [(u/sum1 - v/sum2)**2/(u/n1sq + v/n2sq) \
                 for u, v in zip(vector1, vector2) \
                 if u > 0 and v > 0]
    except ZeroDivisionError:
        ratio = float(sum1)/sum2
        ssq = ratio**2
        terms = [(u/ratio - v)**2/(u/ssq + v) \
                 for u, v in zip(vector1, vector2) \
                 if u > 0 and v > 0]
        
    result = sum(terms)
    return result

def bdm(vector1, vector2):
    r"""Compare numerical set vector1 and vector2
        using the Bhattacharyya distance measure.
        Treating their entries vectors, normalize, and take the dot product.
        Output:
        ts : float
            The BDM test statistic
        """
    assert len(vector1) == len(vector2)
    
    if sum(vector1) == 0 or sum(vector2) == 0:
        return 1. # if they're both empty they're identical

    terms = [u*v for u, v in zip(vector1, vector2)]
    sum1 = sum(vector1)
    sum2 = sum(vector2)
    denominator = sqrt(sum1*sum2)
    result = sqrt(sum(terms))/denominator\
        if denominator else 0.
    return result
    
def anderson_darling(vector1, vector2):
    r"""
    Calculates the AD test statistic between
    vector1 and vector2
    """
    result = 0.
    n_1 = float(sum(vector1))
    n_2 = float(sum(vector2))

    if n_1 == 0 or n_2 == 0:
        return 0.

    factor = 1./(n_1+n_2)
    sigma_j = 0.
    sigma_uj = 0.
    sigma_vj = 0.
    for v_1, v_2 in zip(vector1, vector2):

        if v_1 == 0 and v_2 == 0:
            continue
        term = v_1 + v_2
        
        sigma_uj += v_1
        sigma_vj += v_2
        sigma_j += term
        
        term1 = (1./n_1)*((n_1+n_2)*sigma_uj - n_1*sigma_j)**2
        term2 = (1./n_2)*((n_1+n_2)*sigma_vj - n_2*sigma_j)**2

        denominator = sigma_j *(n_1 + n_2 - sigma_j)
        if denominator == 0:
            continue
        result += term*(term1 + term2)/denominator
        
    result *= factor
    return result

def cramer_von_mises(vector1, vector2):
    r"""
    Compare sequences vector1, vector2 with the Cramer-von-Mises test.
    """
    result = 0.
    if not any(vector1) and not any(vector2):
        return result

    sum1 = float(sum(vector1))
    sum2 = float(sum(vector2))

    for i, u_v in enumerate(zip(vector1, vector2)):
        u_ecdf = sum(vector1[:i])/sum1
        v_ecdf = sum(vector2[:i])/sum2
        result += sum(u_v)*(u_ecdf - v_ecdf)**2
    factor = sum1*sum2/(sum1+sum2)**2
    result *= factor
    return result

def kolmogorov_smirnov(vector1, vector2):
    r"""Calculate the Kolmogorov-Smirnov test statistic
        for two numerical sequences of the same length
    
        Output:
        T : float
    Maximum distance between the cumulative distributions.
    """
    assert len(vector1) == len(vector2)
    
    nbins = len(vector1)
    sum1 = sum(vector1)
    sum2 = sum(vector2)

    if sum1 == 0 and sum2 == 0:
        return 0.

    cdf_diffs = [fabs(sum(vector1[:i])/sum1 - sum(vector2[:i])/sum2)
                 for i in range(nbins)]
    result = max(cdf_diffs)
    return result

def llh_ratio(vector1, vector2):
    r"""
    Calculates the Likelihood ratio between two sequences.
    """
    result = 0.
    if not any(vector1) and not any(vector2):
        return result
    
    sum1 = float(sum(vector1))
    sum2 = float(sum(vector2))

    for u_1, u_2 in zip(vector1, vector2):
        sum12 = u_1 + u_2
        if u_1 == 0 and u_2 == 0:
            result += 1
            continue
        if u_1 == 0:
            result += sum12*log(sum1/(sum1+sum2))
            continue
        if u_2 == 0:
            result += sum12*log(sum2/(sum1+sum2))
            continue
        term1 = sum12*log((1+u_2/u_1)/(1+sum2/sum1))
        term2 = u_2*log((sum2/sum1)*(u_1/u_2))
        result += term1 + term2

    return -2*result

def llh_value(vector1, vector2):
    r"""
    Compare histograms h1, h2 with the log likelihood value test.
    FIXME: This is currently returning inf on occasion.  It should
           never do that. Taking it out of the rotation.
    """
    result = 0.
    if not any(vector1) and not any(vector2):
        return result

    sum1 = float(sum(vector1))
    sum2 = float(sum(vector2))

    for u_1, u_2 in zip(vector1, vector2):
        sum12 = u_1 + u_2
        term1 = log(binom(sum12, u_2))
        term2 = sum12*log(sum1/(sum1 + sum2))
        term3 = u_2*log(sum2/sum1)
        result += term1 + term2 + term3

    return -result







#!/usr/bin/env python3

'''
This example exercises the two sample statistical tests
available from scipy:
* scipy.stats.ttest_ind
* scipy.stats.ks_2samp
* scipy.stats.epps_singleton_2samp
* scipy.stats.mannwhitneyu
* scipy.stats.ranksums
* scipy.stats.wilcoxon
* scipy.stats.kruskal
* scipy.stats.friedmanchisquare
* scipy.stats.brunnermunzel
* scipy.stats.anderson_ksamp
'''

import collections

import numpy
import pylab  # type: ignore[import]
import scipy.stats  # type: ignore[import]

import voka.tools.render

# Low p-value is bad here.
# The null hypothesis H_0 is that the two samples
# were sampled from the same underlying distribution.
# A low p-value implies there's a low probability
# that the difference observed is due to fluctuations
# alone, indicating a systematic and significant difference.
# In an analysis this would be evidence of a signl.
# We effectively want background and background.
# We want no signal.

# Note:
# 1) friedman chisq test fails with
#    'ValueError: Less than 3 levels.  Friedman test not appropriate.'
# 2) wolcoxon fails if the two are distributions are identical.
# 3) anderson_ksamp result for equal distributions:
#    anderson_ksampResult(statistic=-1.3146384545742542,
#                         critical_values=array([0.325,
#                                                1.226,
#                                                1.961,
#                                                2.718,
#                                                3.752,
#                                                4.592,
#                                                6.546]),
#                         significance_level=0.25)
#
# Either way I'm creating a voka.2sample and voka.ksample.

def voka_2sample(sample1, sample2):
    # Checkout OnlineL2_SplitTime2_SPE2itFitEnergy
    # hiccup #1 (AD) ValueError: anderson_ksamp needs more than one distinct observation
    # hiccup #2 (ES) numpy.linalg.LinAlgError: SVD did not converge
    # hiccup #3 (TT) Ttest_indResult(statistic=nan, pvalue=nan)
    # hiccup #4 (MW) ValueError: All numbers are identical in mannwhitneyu
    # hiccup #5 (WP) ValueError: zero_method 'wilcox' and 'pratt' do not work if x - y is zero for all elements
    # hiccup #6 (FC) ValueError: Less than 3 levels.  Friedman test not appropriate.

    result = dict()

    r = scipy.stats.ttest_ind(sample1, sample2)
    result['TTest'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    r = scipy.stats.ks_2samp(sample1, sample2)
    result['KolmogorovSmirnov'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.epps_singleton_2samp(sample1, sample2)
        result['EppsSingleton'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except numpy.linalg.LinAlgError:
        print("    skipping epps_singleton_2samp")

    try:
        r = scipy.stats.mannwhitneyu(sample1, sample2)
        result['MannWhitneyU'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        print("    skipping mannwhitneyu")

    r = scipy.stats.ranksums(sample1, sample2)
    result['Ranksums'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.wilcoxon(sample1, sample2)
        result['Wilcoxon'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        print("    skipping wilcoxon")

    try:
        r = scipy.stats.kruskal(sample1, sample2)
        result['Kruskal'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except:
        print("    skipping kruskal")

#    try:
#        r = scipy.stats.friedmanchisquare(sample1, sample2)
#        result['FriedmanChiSquare'] = {
#            'statistic': r.statistic,
#            'pvalue': r.pvalue
#        }
#    except ValueError:
#        print("    skipping friedmanchisquare")

    r = scipy.stats.brunnermunzel(sample1, sample2)
    result['BrunnerMunzel'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    return result

DENSITY = True

def compare(benchmark_distribution, systematic_distribution, pvalues, density=False):
    # Use the same histogram settings for each.
    _range = (-5,5)
    _bins = 100

    # First histogram to make samples
    benchmark_sample = numpy.histogram(benchmark_distribution,
                                       density=density,
                                       range=_range,
                                       bins=_bins)[0]
    systematic_sample = numpy.histogram(systematic_distribution,
                                        range=_range,
                                        density=density,
                                        bins=_bins)[0]

    # Compare and print the results
    comparison = voka_2sample(benchmark_sample, systematic_sample)
    for k, v in comparison.items():
        pvalues[k].append(v['pvalue'])

if __name__ == '__main__':

    # Tests to perform
    # 1) Poissonian-only differences
    # 2) Normalization differences
    # 3) Mean pull
    # 4) Width differences

    benchmark_width = 1.0
    benchmark_center = 0.0
    benchmark_size = 100000

    widths = numpy.arange(0.5*benchmark_width,
                          1.5*benchmark_width,
                          0.001)

    centers = numpy.arange(-0.5,
                           0.5,
                           0.001)

    sizes = numpy.arange(0.5*benchmark_size,
                         1.5*benchmark_size,
                         10)

    draw_width_idx = int(0.9*len(widths))
    draw_center_idx = int(0.9*len(centers))

    benchmark_distribution = numpy.random.normal(loc=benchmark_center,
                                                 scale=benchmark_width,
                                                 size=benchmark_size)

    # Generate systematic distributions
    print(80*'-')
    figure_number = 1

    pvalues_width_systematic: dict = collections.defaultdict(list)
    for idx, systematic_width in enumerate(widths):
        #print('width = %.4f' % systematic_width)
        systematic_distribution = numpy.random.normal(loc=benchmark_center,
                                                      scale=systematic_width,
                                                      size=benchmark_size)
        compare(benchmark_distribution,
                systematic_distribution,
                pvalues_width_systematic,
                density=DENSITY)

        if idx == draw_width_idx:
            print('width = %.4f' % systematic_width)
            # Use the same histogram settings for each.
            _range = (-5,5)
            _bins = 100

            # First histogram to make samples
            benchmark_sample = numpy.histogram(benchmark_distribution,
                                               density=DENSITY,
                                               range=_range,
                                               bins=_bins)[0]
            systematic_sample = numpy.histogram(systematic_distribution,
                                                density=DENSITY,
                                                range=_range,
                                                bins=_bins)[0]

            pylab.figure(figure_number)
            figure_number += 1
            voka.tools.render.draw_ratio(systematic_sample, benchmark_sample)


    pylab.figure(figure_number)
    figure_number += 1
    for label, pvs in pvalues_width_systematic.items():
        min_pvalue = min(pvs)
        y = [numpy.log10(p) for p in pvs]
        pylab.plot(widths, y, label=label)

    pylab.title("Systematic Width Variation")
    pylab.legend()
    pylab.xlabel('width')
    pylab.ylabel('log(p-value)')

    print(80 * '-')
    pvalues_center_systematic: dict = collections.defaultdict(list)
    for idx, systematic_center in enumerate(centers):
        #print('center = %.4f' % systematic_center)
        systematic_distribution = numpy.random.normal(loc=systematic_center,
                                                      scale=benchmark_width,
                                                      size=benchmark_size)
        compare(benchmark_distribution,
                systematic_distribution,
                pvalues_center_systematic,
                density=DENSITY)

        if idx == draw_center_idx:
            print('center = %.4f' % systematic_center)
            # Use the same histogram settings for each.
            _range = (-5,5)
            _bins = 100

            # First histogram to make samples
            benchmark_sample = numpy.histogram(benchmark_distribution,
                                               density=DENSITY,
                                               range=_range,
                                               bins=_bins)[0]
            systematic_sample = numpy.histogram(systematic_distribution,
                                                density=DENSITY,
                                                range=_range,
                                                bins=_bins)[0]

            pylab.figure(figure_number)
            figure_number += 1
            voka.tools.render.draw_ratio(systematic_sample, benchmark_sample)

    pylab.figure(figure_number)
    figure_number += 1
    for label, pvs in pvalues_center_systematic.items():
        y = [numpy.log10(p) for p in pvs]
        pylab.plot(centers, y, label=label)

    pylab.title("Systematic Center Variation")
    pylab.legend()
    pylab.xlabel('center')
    pylab.ylabel('log(p-value)')

    print(80*'-')
    pylab.figure(figure_number)
    figure_number += 1
    pvalues_size_systematic: dict = collections.defaultdict(list)
    for systematic_size in sizes:
        #print('size = %d' % int(systematic_size))
        systematic_distribution = numpy.random.normal(loc=benchmark_center,
                                                      scale=benchmark_width,
                                                      size=int(systematic_size))
        compare(benchmark_distribution,
                systematic_distribution,
                pvalues_size_systematic,
                density=DENSITY)

    for label, pvs in pvalues_size_systematic.items():
        y = [numpy.log10(p) for p in pvs]
        pylab.plot(sizes, y, label=label)

    pylab.title("Systematic Size Variation")
    pylab.legend()
    pylab.xlabel('size')
    pylab.ylabel('log(p-value)')

    pylab.show()



'''
  Module that contains the function that calculates a Local Outlier Factor.
  https://www.dbs.ifi.lmu.de/Publikationen/Papers/LOF.pdf
'''
import numpy

def distance(vector1, vector2):
    '''
    Euclidean distance.
    '''
    array1 = numpy.array(vector1)
    array2 = numpy.array(vector2)
    return numpy.linalg.norm(array1-array2)

def reach(candidate, k_distance, cluster_member, cluster):

    '''
    k-distance of candidate object p is defined as the distance d(p,o)
    between candidate p and an object o in cluster D such that:
        i)  For at least k objects o' in D/{p} d(p,o') <= d(p,o)
        ii) For at most k-1 objects o' in D/{p} d(p,o') < d(p,o)
    '''
    distances = list()
    for member in cluster:
        d = distance(candidate, member)
        if d > 0:
            distances.append(d)
    distances.sort()
    result = max(distances[:k_distance]) if distances[:k_distance] else 0.
    return max([result, distance(candidate, cluster_member)])

def local_reachability_density(test_point, k_distance, cluster):
    '''
    Local Reachability Density
    '''
    denominator = sum([reach(test_point, k_distance, member, cluster)
                       for member in cluster])
    return len(cluster)/denominator if denominator else 0.

def local_outlier_factor(test_point, k_distance, cluster):
    '''
    Return the LocalOutlierFactor for point 'p' compared
    to collection of reference points in 'D', using k-distance 'k'.
    '''
    ratios = list()
    for member in cluster:
        numerator = local_reachability_density(member,
                                               k_distance,
                                               cluster)

        denominator = local_reachability_density(test_point,
                                                 k_distance,
                                                 cluster)
        if denominator:
            ratios.append(numerator/denominator)

    result = sum(ratios)/float(len(ratios)) if ratios else 0.
    return result



#!/usr/bin/env python3

'''
This example illustrates the failure of traditional statistical
comparisons, such as chi^2, for weighted histograms from
stochastic processes.  Can't rely on the p-value calculated
from standard tools such as scipy even in this simple example.

Histogram the arrival time, expected to be gaussian, of the charge
sampled from a gaussian.
'''

import collections

import numpy
import pylab  # type: ignore[import]

# What's the difference here between voka.metrics.chisq
# and scipy.stats.chisquare ?
# import voka.metrics.chisq
import scipy.stats  # type: ignore[import]

histograms = dict()
N_HISTOGRAMS = 1000
BINS = 50
SIZE = 1000
print("Generating histograms...")
for i in range(N_HISTOGRAMS):
    data = numpy.random.uniform(low=-5, high=5, size=SIZE)
    histograms['Uniform%d' % i] = numpy.histogram(data, bins=BINS)

    data = numpy.random.chisquare(df=5, size=SIZE)
    histograms['ChiSq5%d' % i] = numpy.histogram(data, bins=BINS)

    data = numpy.random.normal(size=SIZE)
    histograms['Gaussian%d' % i] = numpy.histogram(data, bins=BINS)

    data = numpy.random.normal(size=int(SIZE/2), loc=-1, scale=0.25) + \
        numpy.random.normal(size=int(SIZE/2), loc=1, scale=0.25)
    histograms['BiGaussian%d' % i] = numpy.histogram(data, bins=BINS)

    data = numpy.random.exponential(scale=10, size=1000)
    histograms['Exponential%d' % i] = numpy.histogram(data, bins=BINS)

T_dist = collections.defaultdict(list)
#test_stat = voka.metrics.chisq.NormChiSq()
test_stat = scipy.stats.chisquare
print("Comparing histograms...")
for i in range(N_HISTOGRAMS):
    for j in range(i+1, N_HISTOGRAMS):
        h1 = histograms["Uniform%d" % i][0]
        h2 = histograms["Uniform%d" % j][0]
        #T_dist['uniform'].append(test_stat(h1, h2))
        T_dist['uniform'].append(test_stat(h1, h2)[0])

        h1 = histograms["ChiSq5%d" % i][0]
        h2 = histograms["ChiSq5%d" % j][0]
        #T_dist['chisq'].append(test_stat(h1, h2))
        T_dist['chisq'].append(test_stat(h1, h2)[0])

        h1 = histograms["Exponential%d" % i][0]
        h2 = histograms["Exponential%d" % j][0]
        #T_dist['exponential'].append(test_stat(h1, h2))
        T_dist['exponential'].append(test_stat(h1, h2)[0])

        h1 = histograms["Gaussian%d" % i][0]
        h2 = histograms["Gaussian%d" % j][0]
        #T_dist['gaussian'].append(test_stat(h1, h2))
        T_dist['gaussian'].append(test_stat(h1, h2)[0])

        h1 = histograms["BiGaussian%d" % i][0]
        h2 = histograms["BiGaussian%d" % j][0]
        #T_dist['bigaussian'].append(test_stat(h1, h2))
        T_dist['bigaussian'].append(test_stat(h1, h2)[0])

pylab.figure(1)
pylab.subplot(231)
pylab.title(r'$\chi^{2}$ Test Statistic Distribution - Uniform')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')
bin_values, bin_edges, patches = pylab.hist(T_dist['uniform'],
                                            density=True,
                                            range=(0, 200),
                                            bins=100)
x = range(0, int(bin_edges[-1]))
for ndof in range(10, 60, 10):
    rv = scipy.stats.chi2(ndof)
    pylab.plot(x, rv.pdf(x), label='NDOF=%d' % ndof)
pylab.legend()

pylab.subplot(232)
pylab.title(r'$\chi^{2}$ Test Statistic Distribution - ChiSq')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')
bin_values, bin_edges, patches = pylab.hist(T_dist['chisq'],
                                            density=True,
                                            range=(0, 200),
                                            bins=100)
x = range(0, int(bin_edges[-1]))
for ndof in range(10, 60, 10):
    rv = scipy.stats.chi2(ndof)
    pylab.plot(x, rv.pdf(x), label='NDOF=%d' % ndof)
pylab.legend()

pylab.subplot(233)
pylab.title(r'$\chi^{2}$ Test Statistic Distribution - Gaussian')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')
bin_values, bin_edges, patches = pylab.hist(T_dist['gaussian'],
                                            density=True,
                                            range=(0, 200),
                                            bins=100)
x = range(0, int(bin_edges[-1]))
for ndof in range(10, 60, 10):
    rv = scipy.stats.chi2(ndof)
    pylab.plot(x, rv.pdf(x), label='NDOF=%d' % ndof)
pylab.legend()

pylab.subplot(234)
pylab.title(r'$\chi^{2}$ Test Statistic Distribution - Exponential')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')
bin_values, bin_edges, patches = pylab.hist(T_dist['exponential'],
                                            density=True,
                                            range=(0, 200),
                                            bins=100)
x = range(0, int(bin_edges[-1]))
for ndof in range(10, 60, 10):
    rv = scipy.stats.chi2(ndof)
    pylab.plot(x, rv.pdf(x), label='NDOF=%d' % ndof)
pylab.legend()

pylab.subplot(235)
pylab.title(r'$\chi^{2}$ Test Statistic Distribution - BiGaussian')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')
bin_values, bin_edges, patches = pylab.hist(T_dist['bigaussian'],
                                            density=True,
                                            range=(0, 200),
                                            bins=100)
x = range(0, int(bin_edges[-1]))
for ndof in range(10, 60, 10):
    rv = scipy.stats.chi2(ndof)
    pylab.plot(x, rv.pdf(x), label='NDOF=%d' % ndof)
pylab.legend()

pylab.subplot(236)
pylab.title(r'$\chi^{2}$ Test Statistic Distributions')
pylab.xlabel(r'$\chi^{2}$')
pylab.ylabel('Probability Density')

pylab.hist(T_dist['uniform'],
           histtype='step',
           label='Uniform',
           density=True,
           range=(0, 200),
           bins=100)

pylab.hist(T_dist['bigaussian'],
           histtype='step',
           label='BiGaussian',
           density=True,
           range=(0, 200),
           bins=100)

pylab.hist(T_dist['gaussian'],
           histtype='step',
           label='Gaussian',
           density=True,
           range=(0, 200),
           bins=100)

pylab.hist(T_dist['exponential'],
           histtype='step',
           label='Exponential',
           density=True,
           range=(0, 200),
           bins=100)

pylab.hist(T_dist['chisq'],
           histtype='step',
           label='ChisSq',
           density=True,
           range=(0, 200),
           bins=100)

pylab.legend()

pylab.show()



'''
The model module currently consists solely of the Voka class.
'''

import math
import collections
import voka.lof
        
def average_lof(test_sequence, benchmark_sequences):
    
    # this is a list of statistical test points
    reference_cluster = list()
    for idx in range(len(benchmark_sequences)-1):
        for jdx in range(idx, len(benchmark_sequences)):
            s0 = benchmark_sequences[idx]
            s1 = benchmark_sequences[jdx]
            c = voka.metrics.shape_chisq(s0, s1)
            a = voka.metrics.anderson_darling(s0, s1)    
            p = (c,a)
            reference_cluster.append(p)

    test_cluster = list()
    for s1 in benchmark_sequences:
        s0 = test_sequence
        c = voka.metrics.shape_chisq(s0, s1)
        a = voka.metrics.anderson_darling(s0, s1)
        p = (c,a)        
        test_cluster.append(p)

    print('calculating...')
    # now calculate local outlier factors
    k = 3
    lofs = list()
    print(len(test_cluster))
    print(len(reference_cluster))    
    for test_point in test_cluster:
        lofs.append(voka.lof.local_outlier_factor(test_point, k, reference_cluster))
    average_lof = sum(lofs)/float(len(lofs))
    return (average_lof, test_cluster, reference_cluster)
        
        
class Voka:
    '''
    Class to handle determination of the outlier detection thresholds
    for a given set of benchmark samples.
    '''
    def __init__(self):
        # the reference collection is a dictionary containing
        # as values dictionaries with the same structure as the test
        # dictionary.  The keys are arbitrary names for the different
        # sets.
        # Reference {'', {'', []}}
        # Test {'', []}
        self.__reference_collection = dict()
        self.__k = int()
        self.__thresholds = dict()

    def train(self,
              reference_collection,
              k=3,
              tolerance_factor=math.sqrt(2)):
        '''
        Calculate LOF thresholds from the reference set.
        '''

        # there should be a transformation of the reference collection here
        # involving a comparison using statistical tests
        
        self.__reference_collection = reference_collection
        self.__k = k
    
        # we use each one as a test and the others
        # as a benchmark set and determine the
        lof_values = collections.defaultdict(list)
        for test_collection in reference_collection.values():
            # No need to remove the set from itself.
            # Identity should resolve to 0 in each test
            # contributing nothing to the calculation of
            # the average.
            result = self.execute(test_collection)
            assert(result)

            for key, lof in result.items():
                lof_values[key].append(lof)

        self.__thresholds = {histogram_name: tolerance_factor*max(lofs)
                             for histogram_name, lofs in lof_values.items()}
        assert(self.__thresholds)

    def execute(self, test):
        '''
        calculate the thresholds from the benchmark set
        '''
        result = dict()
        for test_key, test_sequence in test.items():

            # pull the reference sequences out of the collection
            reference_sequences = list()
            for ref_set in self.__reference_collection.values():
                if test_key in ref_set:
                    reference_sequences.append(ref_set[test_key])

            # the transformation has to happen here
            # we have a single test_sequence and we have
            # reference sequences
                    
            lof = voka.lof.local_outlier_factor(test_sequence,
                                                self.__k,
                                                reference_sequences)

            result[test_key] = lof
        return result

    def results(self, results):
        '''
        Apply the thresholds determined during training
        and indicate pass/fail.
        '''
        result = dict()
        for key, lof in results.items():
            if key not in self.__thresholds:
                continue
            result[key] = {
                'pass': lof <= self.__thresholds[key],
                'lof': lof,
                'threshold': self.__thresholds[key]
            }
        return result



#!/usr/bin/env python3
'''
Basic example which creates a dummy test dataset and a
benchmark collection.  This example 'trains' on the benchmark
set and displays the results of the comparison.
'''

import numpy

import voka.model

test_data = {
    'Gaussian': numpy.random.normal(size=100),
    'Uniform': numpy.random.uniform(size=100)
}

N_BENCHMARK_COLLECTIONS = 5
benchmark_labels = ['Benchmark_%d' % i for i in range(N_BENCHMARK_COLLECTIONS)]
benchmark_data = {
    benchmark_label:{'Gaussian': numpy.random.normal(size=100),
                     'Uniform': numpy.random.uniform(size=100)}
    for benchmark_label in benchmark_labels
}

# histogram the data
test_histograms = {name:numpy.histogram(data)[0]
                   for name, data in test_data.items()}
benchmark_histograms = dict()
for name, bm_data in benchmark_data.items():
    benchmark_histograms[name] = {n:numpy.histogram(data)[0]
                                  for n, data in bm_data.items()}

voka_test = voka.model.Voka()
voka_test.train(benchmark_histograms)
result = voka_test.execute(test_histograms)
print(result)
print(voka_test.results(result))



#!/usr/bin/env python3
'''
Module that tests the main voka class 'model'
'''

import unittest
import random

import numpy

import voka.model

class TestVoka(unittest.TestCase):
    '''
    Test class that tests the main voka model.
    '''

    def setUp(self):
        dist_mean = 0.
        sigma = 1.

        histogram_names = ["Histogram%d" % i for i in range(100)]
        self.test_hist = dict()
        for name in histogram_names:
            dist = [random.gauss(dist_mean, sigma) for _ in range(1000)]
            self.test_hist[name] = numpy.histogram(dist)[0]

        reference_names = ["ReferenceRun%d" % i for i in range(5)]
        self.reference_collection = {name:dict() for name in reference_names}
        for reference_name in reference_names:
            # For each run generate a set of histograms
            # with the same structure and names as the test histograms

            for name in histogram_names:
                dist = [random.gauss(dist_mean, sigma) for _ in range(1000)]
                self.reference_collection[reference_name][name] =\
                    numpy.histogram(dist)[0]

    def test_voka(self):
        model = voka.model.Voka()
        model.train(self.reference_collection)
        result = model.execute(self.test_hist)
        self.assertTrue(len(result) == 100)

if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sys

from math import sqrt
from math import isnan
import random

import numpy

from voka.histogram import Histogram

class TestHistogram(unittest.TestCase):

    def setUp(self):
        xmin=-5
        xmax=5
        nbins=20
        name='gaussian'
        self.N = 1000
        self.histogram = Histogram(xmin=xmin,
                                   xmax=xmax,
                                   nbins=20,
                                   name='gaussian')

        for _ in range(self.N):
            self.histogram.fill(random.gauss(0., 1.))
        
    def test_fill(self):
        self.assertEqual(sum(self.histogram.bin_values), self.N)
        print(self.histogram.bin_values)
        
    def test_get_state(self):
        state = self.histogram.__getstate__()
        print(state)
        self.assertEqual(state['name'], 'gaussian')            
        self.assertEqual(sum(state['bin_values']), self.N)
        
    def test_set_state(self):
        name = 'possion'
        xmin = 0
        xmax = 10
        nbins = 30
        
        bin_values, _ = numpy.histogram(numpy.random.poisson(5, 1000), bins=nbins, range=(xmin, xmax))
        
        state = {
            "name" : name,
            "xmin" : xmin,
            "xmax" : xmax,
            "overflow" : 0,
            "underflow" : 0,
            "nan_count" : 0,
            "bin_values" : bin_values
            }

        histogram = Histogram(0,1,0,'')
        
        histogram.__setstate__(state)        
        self.assertEqual(histogram.name, name)            
        self.assertEqual(histogram.xmin, xmin)
        self.assertEqual(histogram.xmax, xmax)            
        self.assertEqual(len(histogram.bin_values), nbins)            
        
if __name__ == '__main__':        
    unittest.main()




#!/usr/bin/env python3

'''
This example exercises the two sample statistical tests
available from scipy:
* scipy.stats.ttest_ind
* scipy.stats.ks_2samp
* scipy.stats.anderson_ksamp
* scipy.stats.epps_singleton_2samp
* scipy.stats.mannwhitneyu
* scipy.stats.ranksums
* scipy.stats.wilcoxon
* scipy.stats.kruskal
* scipy.stats.friedmanchisquare
* scipy.stats.brunnermunzel
'''

import collections
import subprocess

import numpy
import scipy.stats  # type: ignore[import]
import pylab  # type: ignore[import]
import matplotlib  # type: ignore[import]

import voka.model
import voka.metrics

matplotlib.use('agg')

def voka_2sample(sample1, sample2):
    print(sample1, sample2)

    # Checkout OnlineL2_SplitTime2_SPE2itFitEnergy
    # hiccup #1 (AD) ValueError: anderson_ksamp needs more than one distinct observation
    # hiccup #2 (ES) numpy.linalg.LinAlgError: SVD did not converge
    # hiccup #3 (TT) Ttest_indResult(statistic=nan, pvalue=nan)
    # hiccup #4 (MW) ValueError: All numbers are identical in mannwhitneyu
    # hiccup #5 (WP) ValueError: zero_method 'wilcox' and 'pratt' do not work if x - y is zero for all elements
    # hiccup #6 (FC) ValueError: Less than 3 levels.  Friedman test not appropriate.

    result = dict()

    r = scipy.stats.ttest_ind(sample1, sample2)
    result['TTest'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    r = scipy.stats.ks_2samp(sample1, sample2)
    result['KolmogorovSmirnov'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.anderson_ksamp([sample1, sample2])
        result['AndersonDarling'] = {
            'statistic': r.statistic,
            'significance_level': r.significance_level
        }
    except ValueError:
        #print("    skipping anderson_ksamp")
        pass

    try:
        r = scipy.stats.epps_singleton_2samp(sample1, sample2)
        result['EppsSingleton'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except numpy.linalg.LinAlgError:
        #print("    skipping epps_singleton_2samp")
        pass

    try:
        r = scipy.stats.mannwhitneyu(sample1, sample2)
        result['MannWhitneyU'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping mannwhitneyu")
        pass

    r = scipy.stats.ranksums(sample1, sample2)
    result['Ranksums'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.wilcoxon(sample1, sample2)
        result['Wilcoxon'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping wilcoxon")
        pass

    try:
        r = scipy.stats.kruskal(sample1, sample2)
        result['Kruskal'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except:
        #print("    skipping kruskal")
        pass

    try:
        r = scipy.stats.friedmanchisquare(sample1, sample2)
        result['FriedmanChiSquare'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping friedmanchisquare")
        pass

    r = scipy.stats.brunnermunzel(sample1, sample2)
    result['BrunnerMunzel'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    return result

# make two samples containing
# 'standard' numpy distributions
locs = numpy.arange(-0.25, 0.25, 0.01)

def gaussian_sample(loc=0.0, size=1000):
    RANGE = (-5,5)
    SCALE = 1.0
    return numpy.histogram(numpy.random.normal(size=size, scale=SCALE, loc=loc),
                           range=RANGE)[0]

unbinned_test_samples = [numpy.random.normal(size=1000, scale=1.0, loc=loc)
                         for loc in locs]

voka_test_samples = [gaussian_sample(loc=loc, size=1000) for loc in locs]
benchmark_samples = [gaussian_sample(size=1000) for _ in range(7)]

results = collections.defaultdict(list)
cmd = "convert -delay 50 -loop 0".split()
for idx, test_sample in enumerate(unbinned_test_samples):

    print("loc = %.2f" % locs[idx])
    benchmark_sample = numpy.random.normal(size=1000)
    voka_2samp_result = voka_2sample(test_sample, benchmark_sample)
    for name, result in voka_2samp_result.items():
        if 'pvalue' in result:
            results[name].append(numpy.log(result['pvalue']))

    lof, test_cluster, reference_cluster = voka.model.average_lof(voka_test_samples[idx], benchmark_samples)
    results['voka'].append(lof)

    pylab.figure()
    marker = 'ro' if lof > 2. else 'go'
    pylab.title('gaussian pull = %f' % locs[idx])
    pylab.xlabel('chisq')
    pylab.ylabel('AD')
    pylab.xlim(0,50)
    pylab.ylim(0,50)
    pylab.plot([p[0] for p in test_cluster],
               [p[1] for p in test_cluster],
               marker)
    pylab.plot([p[0] for p in reference_cluster],
               [p[1] for p in reference_cluster],
               'bo')
    fn = 'illustration_%d.png' % idx
    pylab.savefig(fn)
    cmd.append(fn)
cmd.append('animation.gif')
subprocess.run(cmd)

# 1) now just calculate the metrics for the benchmarks
# and test samples for each pull value
# 2) add the vline for the lof threshold
# 3) animation showing the color change

pylab.figure()
for name, result in results.items():
    if name=='voka':
        continue
    pylab.plot(locs, result, label=name)
pylab.title('Unbinned 2-sample Tests')
pylab.xlabel('gaussian pull')
pylab.axhline(-3.0)
pylab.ylabel('log(p-value)')
pylab.legend()

pylab.figure()
pylab.plot(locs, results['voka'], label='icecube')
pylab.title('LOF-based Method')
pylab.xlabel('gaussian pull')
pylab.axhline(2.0)
pylab.ylabel('<lof>')
pylab.legend()

# #benchmark_x = list()
# #benchmark_y = list()
# #for idx in range(len(benchmark_samples)-1):
# #    for jdx in range(idx+1, len(benchmark_samples)):
# #        benchmark_x = voka.metric.shape_chisq(,)
#
#pylab.show()



#!/usr/bin/env python3

'''
This example illustrates the basic concept of this method,
showing that comparing two histograms using traditional methods
with a p-value threshold is unstable for non-poissonian
distributions.

Voka can be seen as the equivalent of emperically auto-tuning
the p-value threshold on-the-fly.
'''

import numpy
import pylab   # type: ignore[import]

import voka.tools.render
import voka.two_sample

# Generate the histograms
histograms = dict()
N_HISTOGRAMS = 6
for i in range(N_HISTOGRAMS):
    # start with a large Gaussian
    charge_data = list(numpy.random.normal(loc=5.0, scale=0.25, size=10000))

    loc = numpy.random.uniform(3.0, 7.0)
    print(loc)
    random_charge = numpy.random.normal(loc=loc, scale=0.25, size=1000)
    charge_data.extend(random_charge)

    histograms['ChargeHistogram%d' % i] = numpy.histogram(charge_data, bins=100)

# Render the histograms
i=0
for title, histogram in histograms.items():
    pylab.figure(i)
    i=i+1
    bin_values = histogram[0]
    voka.tools.render.draw(bin_values, title)

# Treat the 0-th histogram as the test histogram
# and histograms 1-5 constitute the benchmark ensemble.

# First compare the histograms with traditional statistical tests
test_sample = histograms['ChargeHistogram0'][0]
benchmark_ensemble = [histograms['ChargeHistogram%d' % i][0]
                      for i in range(1,6)]
print(benchmark_ensemble)
for benchmark_sample in benchmark_ensemble:
    result = voka.two_sample.traditional(test_sample, benchmark_sample)
    print(result['KolmogorovSmirnov'])

# Second compare the test histogram using voka.
model = voka.model.Voka()
reference_collection = {
    "Benchmark%d" % idx :
    {"ChargeHistogram":s}
    for idx, s in enumerate(benchmark_ensemble)
}

model.train(reference_collection)
results = model.execute({"ChargeHistogram": test_sample})
result = model.results(results)
print(result)

pylab.show()




import numpy
import scipy.stats  # type: ignore[import]

def traditional(sample1, sample2):

    # hiccup #1 (AD) ValueError: anderson_ksamp needs more than one distinct observation
    # hiccup #2 (ES) numpy.linalg.LinAlgError: SVD did not converge
    # hiccup #3 (TT) Ttest_indResult(statistic=nan, pvalue=nan)
    # hiccup #4 (MW) ValueError: All numbers are identical in mannwhitneyu
    # hiccup #5 (WP) ValueError: zero_method 'wilcox' and 'pratt' do not work if x - y is zero for all elements
    # hiccup #6 (FC) ValueError: Less than 3 levels.  Friedman test not appropriate.

    result = dict()

    r = scipy.stats.ttest_ind(sample1, sample2)
    result['TTest'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    r = scipy.stats.ks_2samp(sample1, sample2)
    result['KolmogorovSmirnov'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.epps_singleton_2samp(sample1, sample2)
        result['EppsSingleton'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except numpy.linalg.LinAlgError:
        #print("    skipping epps_singleton_2samp")
        pass

    try:
        r = scipy.stats.mannwhitneyu(sample1, sample2)
        result['MannWhitneyU'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping mannwhitneyu")
        pass

    r = scipy.stats.ranksums(sample1, sample2)
    result['Ranksums'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    try:
        r = scipy.stats.wilcoxon(sample1, sample2)
        result['Wilcoxon'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping wilcoxon")
        pass

    try:
        r = scipy.stats.kruskal(sample1, sample2)
        result['Kruskal'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except:
        #print("    skipping kruskal")
        pass

    try:
        r = scipy.stats.friedmanchisquare(sample1, sample2)
        result['FriedmanChiSquare'] = {
            'statistic': r.statistic,
            'pvalue': r.pvalue
        }
    except ValueError:
        #print("    skipping friedmanchisquare")
        pass

    r = scipy.stats.brunnermunzel(sample1, sample2)
    result['BrunnerMunzel'] = {
        'statistic': r.statistic,
        'pvalue': r.pvalue
    }

    return result
